@@ -63,7 +63,8 @@
 --  prefix   -> `prefix' `=' string
 --  deriving -> `deriving' `(' ident_1 `,' ... `,' ident_n `)'
 --  parms    -> [verbhs `=>'] `{' parm_1 `,' ... `,' parm_n `}' `->' parm
---  parm     -> [ident_1 [`*' | `-']] verbhs [`&'] [ident_2 [`*'] [`-']]
+--  parm     -> [ident_or_quot_1 [`*' | `-']] verbhs [`&'] [ident_or_quot_2 [`*'] [`-']]
+--  ident_or_quot -> ident | quoths
 --  apath    -> ident
 --            | `*' apath
 --            | apath `.' ident
@@ -87,7 +88,7 @@
 --
 
 module C2HS.CHS (CHSModule(..), CHSFrag(..), CHSHook(..), CHSTrans(..),
-            CHSChangeCase(..), CHSParm(..), CHSArg(..), CHSAccess(..),
+            CHSChangeCase(..), CHSParm(..), CHSMarsh, CHSArg(..), CHSAccess(..),
             CHSAPath(..), CHSPtrType(..),
             loadCHS, dumpCHS, hssuffix, chssuffix, loadCHI, dumpCHI, chisuffix,
             showCHSParm, apathToIdent)
@@ -115,7 +116,6 @@ import C2HS.Version    (version)
 -- friends
 import C2HS.CHS.Lexer  (CHSToken(..), lexCHS, keywordToIdent)
 
-
 -- CHS abstract syntax
 -- -------------------
 
@@ -173,6 +173,8 @@ data CHSHook = CHSImport  Bool                  -- qualified?
                           Position
              | CHSSizeof  Ident                 -- C type
                           Position
+             | CHSAlignof Ident                 -- C type
+                          Position
              | CHSEnum    Ident                 -- C enumeration type
                           (Maybe Ident)         -- Haskell name
                           CHSTrans              -- translation table
@@ -217,6 +219,7 @@ instance Pos CHSHook where
   posOf (CHSContext _ _           pos) = pos
   posOf (CHSType    _             pos) = pos
   posOf (CHSSizeof  _             pos) = pos
+  posOf (CHSAlignof _             pos) = pos
   posOf (CHSEnum    _ _ _ _ _     pos) = pos
   posOf (CHSEnumDefine _ _ _      pos) = pos
   posOf (CHSCall    _ _ _ _       pos) = pos
@@ -237,6 +240,8 @@ instance Eq CHSHook where
     ide1 == ide2
   (CHSSizeof ide1              _) == (CHSSizeof ide2              _) =
     ide1 == ide2
+  (CHSAlignof ide1             _) == (CHSAlignof ide2             _) =
+    ide1 == ide2    
   (CHSEnum ide1 oalias1 _ _ _  _) == (CHSEnum ide2 oalias2 _ _ _  _) =
     oalias1 == oalias2 && ide1 == ide2
   (CHSEnumDefine ide1 _ _      _) == (CHSEnumDefine ide2 _ _      _) =
@@ -266,15 +271,17 @@ data CHSChangeCase = CHSSameCase
                    | CHSDownCase
                    deriving Eq
 
+-- | marshaller consists of a function name or verbatim Haskell code
+--   and flag indicating whether it has to be executed in the IO monad
+--
+type CHSMarsh = Maybe (Either Ident String, CHSArg)
+
 -- | marshalling descriptor for function hooks
 --
--- * a marshaller consists of a function name and flag indicating whether it
---   has to be executed in the IO monad
---
-data CHSParm = CHSParm (Maybe (Ident, CHSArg))  -- "in" marshaller
-                       String                   -- Haskell type
-                       Bool                     -- C repr: two values?
-                       (Maybe (Ident, CHSArg))  -- "out" marshaller
+data CHSParm = CHSParm CHSMarsh  -- "in" marshaller
+                       String    -- Haskell type
+                       Bool      -- C repr: two values?
+                       CHSMarsh  -- "out" marshaller
                        Position
 
 -- | kinds of arguments in function hooks
@@ -478,6 +485,9 @@ showCHSHook (CHSType ide _) =
 showCHSHook (CHSSizeof ide _) =
     showString "sizeof "
   . showCHSIdent ide
+showCHSHook (CHSAlignof   ide _) =
+   showString "alignof "
+  . showCHSIdent ide
 showCHSHook (CHSEnum ide oalias trans oprefix derive _) =
     showString "enum "
   . showIdAlias ide oalias
@@ -574,12 +584,15 @@ showCHSParm (CHSParm oimMarsh hsTyStr twoCVals oomMarsh _)  =
   . showOMarsh oomMarsh
   where
     showOMarsh Nothing               = id
-    showOMarsh (Just (ide, argKind)) =   showCHSIdent ide
-                                       . (case argKind of
-                                           CHSValArg    -> id
-                                           CHSIOArg     -> showString "*"
-                                           CHSVoidArg   -> showString "-"
-                                           CHSIOVoidArg -> showString "*-")
+    showOMarsh (Just (body, argKind)) =   showMarshBody body
+                                        . (case argKind of
+                                             CHSValArg    -> id
+                                             CHSIOArg     -> showString "*"
+                                             CHSVoidArg   -> showString "-"
+                                             CHSIOVoidArg -> showString "*-")
+    --
+    showMarshBody (Left ide) = showCHSIdent ide
+    showMarshBody (Right str) = showChar '|' . showString str . showChar '|'
     --
     showHsVerb str = showChar '`' . showString str . showChar '\''
 
@@ -775,6 +788,7 @@ parseFrags tokens  = do
     parseFrags0 (CHSTokContext pos  :toks) = parseContext pos        toks
     parseFrags0 (CHSTokType    pos  :toks) = parseType    pos        toks
     parseFrags0 (CHSTokSizeof  pos  :toks) = parseSizeof  pos        toks
+    parseFrags0 (CHSTokAlignof pos  :toks) = parseAlignof pos        toks
     parseFrags0 (CHSTokEnum    pos  :toks) = parseEnum    pos        toks
     parseFrags0 (CHSTokCall    pos  :toks) = parseCall    pos        toks
     parseFrags0 (CHSTokFun     pos  :toks) = parseFun     pos        toks
@@ -862,6 +876,14 @@ parseSizeof pos (CHSTokIdent _ ide:toks) =
     return $ CHSHook (CHSSizeof ide pos) : frags
 parseSizeof _ toks = syntaxError toks
 
+parseAlignof :: Position -> [CHSToken] -> CST s [CHSFrag]
+parseAlignof pos (CHSTokIdent _ ide:toks) =
+  do
+    toks' <- parseEndHook toks
+    frags <- parseFrags toks'
+    return $ CHSHook (CHSAlignof ide pos) : frags
+parseAlignof _ toks = syntaxError toks
+
 parseEnum :: Position -> [CHSToken] -> CST s [CHSFrag]
 
 -- {#enum define hsid {alias_1,...,alias_n}  [deriving (clid_1,...,clid_n)] #}
@@ -977,18 +999,27 @@ parseParm toks =
     (oomMarsh, toks'3) <- parseOptMarsh toks'2
     return (CHSParm oimMarsh hsTyStr twoCVals oomMarsh pos, toks'3)
   where
-    parseOptMarsh :: [CHSToken] -> CST s (Maybe (Ident, CHSArg), [CHSToken])
-    parseOptMarsh (CHSTokIdent _ ide:CHSTokStar _ :CHSTokMinus _:toks') =
-      return (Just (ide, CHSIOVoidArg) , toks')
-    parseOptMarsh (CHSTokIdent _ ide:CHSTokStar _ :toks') =
-      return (Just (ide, CHSIOArg) , toks')
-    parseOptMarsh (CHSTokIdent _ ide:CHSTokMinus _:toks') =
-      return (Just (ide, CHSVoidArg), toks')
-    parseOptMarsh (CHSTokIdent _ ide              :toks') =
-      return (Just (ide, CHSValArg) , toks')
-    parseOptMarsh toks'                                   =
+    parseOptMarsh :: [CHSToken] -> CST s (CHSMarsh, [CHSToken])
+    parseOptMarsh (CHSTokIdent _ ide:toks') =
+      do
+        (marshType, toks'2) <- parseOptMarshType toks'
+        return (Just (Left ide, marshType), toks'2)
+    parseOptMarsh (CHSTokHSQuot _ str:toks') =
+      do
+        (marshType, toks'2) <- parseOptMarshType toks'
+        return (Just (Right str, marshType), toks'2)
+    parseOptMarsh toks'                     =
       return (Nothing, toks')
 
+    parseOptMarshType (CHSTokStar _ :CHSTokMinus _:toks') =
+      return (CHSIOVoidArg , toks')
+    parseOptMarshType (CHSTokStar _ :toks') =
+      return (CHSIOArg , toks')
+    parseOptMarshType (CHSTokMinus _:toks') =
+      return (CHSVoidArg, toks')
+    parseOptMarshType toks' =
+      return (CHSValArg, toks')
+
 parseField :: Position -> CHSAccess -> [CHSToken] -> CST s [CHSFrag]
 parseField pos access toks =
   do
