@@ -87,7 +87,6 @@
 --    transfers control to the following binding-hook lexer:
 --
 --      ident       -> letter (letter | digit | `\'')*
---                   | `\'' letter (letter | digit)* `\''
 --      reservedid  -> `as' | `call' | `class' | `context' | `deriving'
 --                   | `enum' | `foreign' | `fun' | `get' | `lib'
 --                   | `downcaseFirstLetter'
@@ -98,14 +97,18 @@
 --      reservedsym -> `{#' | `#}' | `{' | `}' | `,' | `.' | `->' | `='
 --                   | `=>' | '-' | `*' | `&' | `^'
 --      string      -> `"' instr* `"'
---      verbhs      -> `\`' instr* `\''
+--      verbhs      -> `\`' inhsverb* `\''
+--      quoths      -> `\'' inhsverb* `\''
 --      instr       -> ` '..`\127' \\ `"'
+--      inhsverb    -> ` '..`\127' \\ `\''
 --      comment     -> `--' (any \\ `\n')* `\n'
 --
 --    Control characters, white space, and comments are discarded in the
 --    binding-hook lexer.  Nested comments are not allowed in a binding hook.
 --    Identifiers can be enclosed in single quotes to avoid collision with
---    C->Haskell keywords.
+--    C->Haskell keywords, and in this case quoted part could also be followed by
+--    what is assumed to be a valid Haskell code, which would be transferred in the
+--    output file verbatim.
 --
 --  * In the binding-hook lexer, the lexeme `#}' transfers control back to the
 --    base lexer.  An occurence of the lexeme `{#' inside the binding-hook
@@ -128,8 +131,7 @@
 --    is accepted by the C lexer.  In the case of Haskell identifiers, a
 --    confusion between variable and constructor identifiers will be noted by
 --    the Haskell compiler translating the code generated by c2hs.  Moreover,
---    identifiers can be enclosed in single quotes to avoid collision with
---    C->Haskell keywords, but those may not contain apostrophes.
+--    identifiers that are inside quoted parts (see above) may not contain apostrophes.
 --
 --  * Any line starting with the character `#' is regarded to be a C
 --    preprocessor directive.  With the exception of `#c' and `#endc', which
@@ -223,6 +225,7 @@ data CHSToken = CHSTokArrow   Position          -- `->'
               | CHSTokQualif  Position          -- `qualified'
               | CHSTokSet     Position          -- `set'
               | CHSTokSizeof  Position          -- `sizeof'
+              | CHSTokAlignof Position          -- `alignof'
               | CHSTokStable  Position          -- `stable'
               | CHSTokType    Position          -- `type'
               | CHSTok_2Case  Position          -- `underscoreToCase'
@@ -231,6 +234,7 @@ data CHSToken = CHSTokArrow   Position          -- `->'
               | CHSTokWith    Position          -- `with'
               | CHSTokString  Position String   -- string
               | CHSTokHSVerb  Position String   -- verbatim Haskell (`...')
+              | CHSTokHSQuot  Position String   -- quoted Haskell ('...')
               | CHSTokIdent   Position Ident    -- identifier
               | CHSTokHaskell Position String   -- verbatim Haskell code
               | CHSTokCPP     Position String   -- pre-processor directive
@@ -273,6 +277,7 @@ instance Pos CHSToken where
   posOf (CHSTokQualif  pos  ) = pos
   posOf (CHSTokSet     pos  ) = pos
   posOf (CHSTokSizeof  pos  ) = pos
+  posOf (CHSTokAlignof pos  ) = pos
   posOf (CHSTokStable  pos  ) = pos
   posOf (CHSTokType    pos  ) = pos
   posOf (CHSTok_2Case  pos  ) = pos
@@ -281,6 +286,7 @@ instance Pos CHSToken where
   posOf (CHSTokWith    pos  ) = pos
   posOf (CHSTokString  pos _) = pos
   posOf (CHSTokHSVerb  pos _) = pos
+  posOf (CHSTokHSQuot  pos _) = pos
   posOf (CHSTokIdent   pos _) = pos
   posOf (CHSTokHaskell pos _) = pos
   posOf (CHSTokCPP     pos _) = pos
@@ -323,6 +329,7 @@ instance Eq CHSToken where
   (CHSTokQualif   _  ) == (CHSTokQualif   _  ) = True
   (CHSTokSet      _  ) == (CHSTokSet      _  ) = True
   (CHSTokSizeof   _  ) == (CHSTokSizeof   _  ) = True
+  (CHSTokAlignof  _  ) == (CHSTokAlignof  _  ) = True
   (CHSTokStable   _  ) == (CHSTokStable   _  ) = True
   (CHSTokType     _  ) == (CHSTokType     _  ) = True
   (CHSTok_2Case   _  ) == (CHSTok_2Case   _  ) = True
@@ -331,6 +338,7 @@ instance Eq CHSToken where
   (CHSTokWith     _  ) == (CHSTokWith     _  ) = True
   (CHSTokString   _ _) == (CHSTokString   _ _) = True
   (CHSTokHSVerb   _ _) == (CHSTokHSVerb   _ _) = True
+  (CHSTokHSQuot   _ _) == (CHSTokHSQuot   _ _) = True
   (CHSTokIdent    _ _) == (CHSTokIdent    _ _) = True
   (CHSTokHaskell  _ _) == (CHSTokHaskell  _ _) = True
   (CHSTokCPP      _ _) == (CHSTokCPP      _ _) = True
@@ -374,6 +382,7 @@ instance Show CHSToken where
   showsPrec _ (CHSTokQualif  _  ) = showString "qualified"
   showsPrec _ (CHSTokSet     _  ) = showString "set"
   showsPrec _ (CHSTokSizeof  _  ) = showString "sizeof"
+  showsPrec _ (CHSTokAlignof _  ) = showString "alignof"
   showsPrec _ (CHSTokStable  _  ) = showString "stable"
   showsPrec _ (CHSTokType    _  ) = showString "type"
   showsPrec _ (CHSTok_2Case  _  ) = showString "underscoreToCase"
@@ -382,6 +391,7 @@ instance Show CHSToken where
   showsPrec _ (CHSTokWith    _  ) = showString "with"
   showsPrec _ (CHSTokString  _ s) = showString ("\"" ++ s ++ "\"")
   showsPrec _ (CHSTokHSVerb  _ s) = showString ("`" ++ s ++ "'")
+  showsPrec _ (CHSTokHSQuot  _ s) = showString ("'" ++ s ++ "'")
   showsPrec _ (CHSTokIdent   _ i) = (showString . identToString) i
   showsPrec _ (CHSTokHaskell _ s) = showString s
   showsPrec _ (CHSTokCPP     _ s) = showString s
@@ -627,6 +637,7 @@ bhLexer  =      identOrKW
            >||< symbol
            >||< strlit
            >||< hsverb
+           >||< hsquot
            >||< whitespace
            >||< endOfHook
            >||< string "--" +> anyButNL`star` char '\n'   -- comment
@@ -672,10 +683,6 @@ identOrKW  =
        -- identifier or keyword
        (letter +> (letter >|< digit >|< char '\'')`star` epsilon
        `lexactionName` \cs pos name -> (idkwtok $!pos) cs name)
-  >||< -- identifier in single quotes
-       (char '\'' +> letter +> (letter >|< digit)`star` char '\''
-       `lexactionName` \cs pos name -> (mkid $!pos) cs name)
-       -- NB: quotes are removed by lexemeToIdent
   where
     idkwtok pos "as"               _    = CHSTokAs      pos
     idkwtok pos "call"             _    = CHSTokCall    pos
@@ -697,6 +704,7 @@ identOrKW  =
     idkwtok pos "qualified"        _    = CHSTokQualif  pos
     idkwtok pos "set"              _    = CHSTokSet     pos
     idkwtok pos "sizeof"           _    = CHSTokSizeof  pos
+    idkwtok pos "alignof"          _    = CHSTokAlignof pos
     idkwtok pos "stable"           _    = CHSTokStable  pos
     idkwtok pos "type"             _    = CHSTokType    pos
     idkwtok pos "underscoreToCase" _    = CHSTok_2Case  pos
@@ -730,6 +738,7 @@ keywordToIdent tok =
     CHSTokQualif  pos -> mkid pos "qualified"
     CHSTokSet     pos -> mkid pos "set"
     CHSTokSizeof  pos -> mkid pos "sizeof"
+    CHSTokAlignof pos -> mkid pos "alignof"
     CHSTokStable  pos -> mkid pos "stable"
     CHSTokType    pos -> mkid pos "type"
     CHSTok_2Case  pos -> mkid pos "underscoreToCase"
@@ -770,14 +779,20 @@ hsverb :: CHSLexer
 hsverb  = char '`' +> inhsverb`star` char '\''
           `lexaction` \cs pos -> Just (CHSTokHSVerb pos (init . tail $ cs))
 
+-- | quoted code
+--
+hsquot :: CHSLexer
+hsquot  = char '\'' +> inhsverb`star` char '\''
+          `lexaction` \cs pos -> Just (CHSTokHSQuot pos (init . tail $ cs))
+
 
 -- | regular expressions
 --
 letter, digit, instr, inhsverb :: Regexp s t
-letter   = alt ['a'..'z'] >|< alt ['A'..'Z'] >|< char '_'
-digit    = alt ['0'..'9']
-instr    = alt ([' '..'\255'] \\ "\"\\")
-inhsverb = alt ([' '..'\127'] \\ "\'")
+letter    = alt ['a'..'z'] >|< alt ['A'..'Z'] >|< char '_'
+digit     = alt ['0'..'9']
+instr     = alt ([' '..'\255'] \\ "\"\\")
+inhsverb  = alt ([' '..'\127'] \\ "'")
 
 -- | character sets
 --
